<?php

namespace Drupal\orange_dam\Commands;

use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueWorkerManager;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\orange_dam\OrangeDamApi;
use Drupal\orange_dam\OrangeDamConfigurationManager;
use Drupal\orange_dam\OrangeDamContent;
use Drupal\orange_dam\OrangeDamMigrationDataManager;
use Drupal\orange_dam\OrangeDamQueueDataManager;
use Drush\Commands\DrushCommands;

/**
 * Builds Orange DAM related drush commands.
 */
class OrangeDamCommands extends DrushCommands {

  use StringTranslationTrait;

  /**
   * The maximum length of a title in the command output.
   */
  public const MAX_TITLE_CHARACTER_LENGTH = 80;

  /**
   * The maximum length of a field name in the command output.
   */
  public const MAX_FIELD_CHARACTER_LENGTH = 80;

  /**
   * The limit to use when migrating a single item.
   */
  public const SINGLE_ITEM_MIGRATION_LIMIT = 1;

  /**
   * Builds the drush commands.
   */
  public function __construct(
    protected OrangeDamQueueDataManager $orangeDamUpdatedDataManager,
    protected OrangeDamMigrationDataManager $orangeDamMigrationDataManager,
    protected OrangeDamApi $orangeDamApi,
    protected OrangeDamConfigurationManager $orangeDamConfigurationManager,
    protected QueueFactory $queueFactory,
    protected QueueWorkerManager $queueManager,
    protected Connection $databaseConnection,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected LoggerChannelFactoryInterface $loggerChannelFactory,
  ) {}

  /**
   * Sync deletions and queue all updated items since the last sync.
   *
   * @command orange-dam:sync-updates
   * @aliases od-su
   *
   * @usage orange-dam:su
   *   Sync deletions and queue all updated items since the last sync.
   */
  public function syncUpdates() {
    $this->orangeDamUpdatedDataManager->syncUpdates();
    $this->displayMigrationQueueListing();
  }

  /**
   * Queue a single content item to be updated from Orange DAM.
   *
   * @param string $system_identifier
   *   The system identifier of the item in Orange DAM. Example: EX1STO431061.
   * @param array $options
   *   Command options.
   *
   * @command orange-dam:queue-content-item
   * @aliases od-qci
   *
   * @option disable-duplicate-check
   *   Disable duplicate item detection.
   *
   * @usage orange-dam:queue-content-item <system_identifier>
   *   Queue an update for a single item from Orange DAM.
   * @usage orange-dam:queue-content-item <system_identifier> --disable-duplicate-check
   *   Queue an update for a single item without duplicate checks.
   */
  public function queueContentItem(
    string $system_identifier,
    array $options = [
      'disable-duplicate-check' => FALSE,
    ],
  ) {
    try {
      OrangeDamContent::validateSystemIdentifier($system_identifier);
    }
    catch (\Exception $e) {
      $this->loggerChannelFactory
        ->get(OrangeDamQueueDataManager::LOGGER_CHANNEL)
        ->error($e->getMessage());
      return;
    }
    // Retrieve the item.
    $this->orangeDamUpdatedDataManager->updateSingleItem($system_identifier);
  }

  /**
   * Queue and migrate a single content item from Orange DAM.
   *
   * @param string $system_identifier
   *   The system identifier of the item in Orange DAM. Example: EX1STO431061.
   * @param array $options
   *   Command options.
   *
   * @command orange-dam:queue-migrate-content-item
   * @aliases od-qmci
   *
   * @option disable-duplicate-check
   *   Disable duplicate item detection.
   *
   * @usage orange-dam:queue-migrate-content-item <system_identifier>
   *   Queue an update for a single item from Orange DAM.
   * @usage orange-dam:queue-migrate-content-item <system_identifier> --disable-duplicate-check
   *   Queue an update for a single item without duplicate checks.
   */
  public function queueAndMigrateContentItem(
    string $system_identifier,
    array $options = [
      'disable-duplicate-check' => FALSE,
    ],
  ) {
    try {
      OrangeDamContent::validateSystemIdentifier($system_identifier);
    }
    catch (\Exception $e) {
      $this->loggerChannelFactory
        ->get(OrangeDamQueueDataManager::LOGGER_CHANNEL)
        ->error($e->getMessage());
      return;
    }
    // Catch a failure to retrieve the item.
    if (!$item = $this->orangeDamUpdatedDataManager->updateSingleItem($system_identifier)) {
      return;
    }
    // Run the migrations.
    $migration = $this->orangeDamMigrationDataManager->getItemMigration($item);
    $migrations = $migration ? [$migration->id()] : [];
    $results = $this->orangeDamMigrationDataManager->runMigrations(
      static::SINGLE_ITEM_MIGRATION_LIMIT,
      [$system_identifier],
      $migrations,
      (bool) $options['disable-duplicate-check'],
    );
    $this->displayMigrationResults($results);
  }

  /**
   * Queue all content updated since a given time.
   *
   * @param string $time
   *   Full date time string in UTC of any PHP parsable format, such as
   *   "1/25/2023 09:47:01".
   *
   * @command orange-dam:queue-content-updated-since-time
   * @aliases od-qcust
   *
   * @usage orange-dam:queue-content-updated-since-time <php parseable date time string>
   *   Queue all content updated since a given time.
   */
  public function queueContentUpdatedSinceTime(string $time) {
    $time = $this->createDateTime($time);
    $this->orangeDamUpdatedDataManager->queueUpdatedItems($time);
    $this->displayMigrationQueueListing([OrangeDamMigrationDataManager::MIGRATION_TAG_CONTENT]);
  }

  /**
   * Queue all content updated since the last sync.
   *
   * @command orange-dam:queue-content-updated-since-last-sync
   * @aliases od-qcusls
   *
   * @usage orange-dam:queue-content-updated-since-last-sync
   *   Queue all content updated since the last sync.
   */
  public function queueContentUpdatedSinceLastSync() {
    $this->orangeDamUpdatedDataManager->queueUpdatedItems();
    $this->displayMigrationQueueListing([OrangeDamMigrationDataManager::MIGRATION_TAG_CONTENT]);
  }

  /**
   * Queue all content of a given type for syncing.
   *
   * @param array $options
   *   The command options.
   *
   * @command orange-dam:queue-content-by-type
   * @aliases od-qcbt
   *
   * @option type
   *   Allows for the item type to be specified.
   *
   * @usage orange-dam:queue-content-by-type
   *   Queue all content for a given content type.
   */
  public function queueContentByType(array $options = [
    'type' => '',
  ]) {
    $content_type = $options['type'];
    if ($content_type) {
      if (!$this->orangeDamUpdatedDataManager->validateContentType($content_type)) {
        $this->io()->error(dt(
          'Content type :type is not a valid content type.', [
            ':type' => $content_type,
          ]));
        return;
      }
    }
    else {
      $content_types = $this->orangeDamConfigurationManager->getContentTypeNames();
      $choice = $this->io()->choice('Content type to sync from Orange DAM:', $content_types);
      $content_type = $content_types[$choice];
    }
    $this->orangeDamUpdatedDataManager->queueContentByType([$content_type]);
    $this->displayMigrationQueueListing([OrangeDamMigrationDataManager::MIGRATION_TAG_CONTENT]);
  }

  /**
   * Queue all content for syncing.
   *
   * @command orange-dam:queue-content-all
   * @aliases od-qca
   *
   * @option limit
   *   Allows for a limit to be specified for each item type.
   *
   * @usage orange-dam:queue-content-all
   *   Queue all content for syncing.
   */
  public function queueContentAll(array $options = [
    'limit' => '',
  ]) {
    $limit = $options['limit'] ?: 0;
    $this->orangeDamUpdatedDataManager->queueContentByType(
      $this->orangeDamConfigurationManager->getContentTypeNames(),
      (int) $limit,
    );
    $this->displayMigrationQueueListing();
  }

  /**
   * Queue updated Orange DAM keywords since the last sync.
   *
   * @command orange-dam:queue-keywords-updated-since-last-sync
   * @aliases od-quksls
   *
   * @usage orange-dam:queue-keywords-updated-since-last-sync
   *   Queue all updated Orange DAM keywords since the last sync.
   */
  public function queueKeywordsUpdatedSinceLastSync() {
    $this->orangeDamUpdatedDataManager->queueUpdatedKeywords();
    $this->displayMigrationQueueListing([OrangeDamMigrationDataManager::MIGRATION_TAG_TAXONOMY]);
  }

  /**
   * Queue updated Orange DAM keywords since a given time.
   *
   * @param string $time
   *   Full date time string in UTC of any PHP parsable format, such as
   *   "1/25/2023 09:47:01".
   *
   * @command orange-dam:queue-keywords-updated-since-time
   * @aliases od-qukst
   *
   * @usage orange-dam:queue-keywords-updated-since-time
   *   Queue all updated Orange DAM keywords since a given time.
   */
  public function queueKeywordsUpdatedSinceTime(string $time) {
    $time = $this->createDateTime($time);
    $this->orangeDamUpdatedDataManager->queueUpdatedKeywords($time);
    $this->displayMigrationQueueListing([OrangeDamMigrationDataManager::MIGRATION_TAG_TAXONOMY]);
  }

  /**
   * Queue Orange DAM keywords by type.
   *
   * @command orange-dam:queue-keywords-by-type
   * @aliases od-qkbt
   *
   * @usage orange-dam:queue-keywords-by-type --type="Example Type"
   *   Queue all updated Orange DAM keywords of a given type.
   * @usage orange-dam:queue-keywords-by-type
   *   Queue all updated Orange DAM keywords of a given type from the prompt.
   */
  public function queueKeywordsByType($options = [
    'type' => '',
  ]) {
    $tag_type = $options['type'];
    if ($tag_type) {
      if (!$this->orangeDamUpdatedDataManager->validateKeywordType($tag_type)) {
        $this->io()->error(dt(
          'Keyword type :type is not a valid keyword type.', [
            ':type' => $tag_type,
          ]));
        return;
      }
    }
    else {
      $tag_types = $this->orangeDamConfigurationManager->getKeywordTypes();
      $choice = $this->io()->choice('Keyword type to sync from Orange DAM:', $tag_types);
      $tag_type = $tag_types[$choice];
    }
    $this->orangeDamUpdatedDataManager->queueKeywordByType([$tag_type]);
    $this->displayMigrationQueueListing([OrangeDamMigrationDataManager::MIGRATION_TAG_TAXONOMY]);
    $this->displayUnregisteredQueueCount();
  }

  /**
   * Queue all keywords for migrations.
   *
   * @command orange-dam:queue-keywords-all
   * @aliases od-qka
   *
   * @usage orange-dam:queue-keywords-all
   *   Queue all content for migrations.
   */
  public function queueKeywordsAll() {
    $this->orangeDamUpdatedDataManager->queueKeywordByType([]);
    $this->displayMigrationQueueListing([OrangeDamMigrationDataManager::MIGRATION_TAG_TAXONOMY]);
  }

  /**
   * Queue all content deletions since a given time.
   *
   * @param string $time
   *   Full date time string in UTC of any PHP parsable format, such as
   *   "1/25/2023 09:47:01".
   *
   * @command orange-dam:queue-deleted-content-since-time
   * @aliases od-qdcst
   *
   * @usage orange-dam:queue-deleted-content-since-time 1/23/2023
   *   Queue all content deletions since a given time.
   */
  public function queueDeletedContentSinceTime(string $time) {
    $time = $this->createDateTime($time);
    $this->orangeDamUpdatedDataManager->queueContentDeletions($time);
  }

  /**
   * Queue all content deletions since the last sync.
   *
   * @command orange-dam:queue-deleted-content-since-last-sync
   * @aliases od-qdcsls
   *
   * @usage orange-dam:queue-deleted-content-since-last-sync
   *   Queue all content deletions since the last sync.
   */
  public function queueDeletionsContentSinceLastSync() {
    $this->orangeDamUpdatedDataManager->queueContentDeletions();
  }

  /**
   * Queue keywords deleted since the given time.
   *
   * @command orange-dam:queue-deleted-keywords-since-time
   * @aliases od-qdkst
   *
   * @usage orange-dam:queue-deleted-keywords-since-time 1/23/2023
   *   Queue keyword deletions since the given time.
   */
  public function queueDeletedKeywordsSinceTime(string $time) {
    $time = $this->createDateTime($time);
    $this->orangeDamUpdatedDataManager->queueKeywordDeletions($time);
  }

  /**
   * Queue keywords deleted since the last sync.
   *
   * @command orange-dam:queue-deleted-keywords-since-last-sync
   * @aliases od-qdksls
   *
   * @usage orange-dam:queue-deleted-keywords-since-last-sync
   *   Queue all keyword deletions since the last sync.
   */
  public function queueDeletedKeywordsSinceLastSync() {
    $this->orangeDamUpdatedDataManager->queueKeywordDeletions();
  }

  /**
   * Run all Orange DAM migrations.
   *
   * @command orange-dam:migration-run-all
   * @aliases od-mra
   *
   * @option limit
   *   The maximum number of items to migrate.
   * @option disable-duplicate-check
   *   Disable duplicate item checking. Useful when importing large data sets
   *   at a time. This option is a flag and does not require a value.
   *   Example orange-dam:migration-run-all --disable-duplicate-check
   *
   * @usage orange-dam:migration-run-all
   *   Run the Orange DAM migrations with the default limit.
   * @usage orange-dam:migration-run-all --limit=123
   *   Run the Orange DAM migrations with a custom limit.
   */
  public function migrationRunAll(array $options = [
    'limit' => '',
    'disable-duplicate-check' => FALSE,
  ]) {
    $limit = $options['limit'] ?: OrangeDamMigrationDataManager::MIGRATION_LIMIT;
    $this->loggerChannelFactory->get(OrangeDamMigrationDataManager::LOGGER_CHANNEL)->notice(dt(
      'Running all Orange DAM related migrations.'
    ));
    $results = $this->orangeDamMigrationDataManager->runMigrations(
      $limit,
      [],
      [],
      (bool) $options['disable-duplicate-check'],
    );
    $this->displayMigrationResults($results);
    $this->displayMigrationQueueListing();
  }

  /**
   * Run a specific Orange DAM migration.
   *
   * @param array $options
   *   The command options.
   *
   * @command orange-dam:migration-run
   * @aliases od-mr
   *
   * @option migration
   *   Allows for the migration to be specified.
   * @option limit
   *   The maximum number of items to migrate.
   * @option idlist
   *   The comma separated source ID of the items to migrate.
   * @option disable-duplicate-check
   *   Disable duplicate item checking. Useful when importing large data sets
   *   at a time. This option is a flag and does not require a value.
   *   Example orange-dam:migration-run-all --disable-duplicate-check
   *
   * @usage orange-dam:migration-run --limit=123
   *   Run a specified Orange DAM migration with the limit of 123.
   * @usage orange-dam:migration-run --migration=example_migration
   *   Run the specified Orange DAM migration with no limit.
   * @usage orange-dam:migration-run --idlist=1,2,3
   *   Run the specified Orange DAM migration for the specified items.
   */
  public function migrationRun(array $options = [
    'migration' => '',
    'limit' => '',
    'idlist' => '',
    'disable-duplicate-check' => FALSE,
  ]) {
    $migrations = $this->orangeDamMigrationDataManager->getMigrationListing();
    // Allow for a user supplied migration option.
    $migration_id = $options['migration'];
    if ($migration_id) {
      if (!$this->orangeDamMigrationDataManager->validateMigrationId($migration_id)) {
        $this->io()->error(dt(
          'The migration ID :migration is not a valid migration.', [
            ':migration' => $migration_id,
          ]));
        return;
      }
    }
    // Allow the user to select the migration interactively.
    else {
      // Prompt the user for a selection and collect the data.
      $migration_id = $this->io()->choice(
        'Migration to run:',
        array_combine(array_keys($migrations), array_keys($migrations)),
      );
    }
    // Inform the user of the action being taken.
    $this->loggerChannelFactory
      ->get(OrangeDamMigrationDataManager::LOGGER_CHANNEL)
      ->notice(dt(
        'Running the :migration migration from queued data.', [
          ':migration' => $migrations[$migration_id],
        ]));
    // Prepare the ID list.
    $idlist = $options['idlist'] ? explode(',', $options['idlist']) : [];
    // Run the migration.
    $results = $this->orangeDamMigrationDataManager->runMigrations(
      $options['limit'],
      $idlist,
      [$migration_id],
      (bool) $options['disable-duplicate-check'],
    );
    // Display the results.
    $this->displayMigrationResults($results);
    $this->displayMigrationQueueListing();
  }

  /**
   * Clear all Orange DAM migration queues.
   *
   * @command orange-dam:migration-clear-queue-all
   * @aliases od-mcqa
   *
   * @usage orange-dam:migration-clear-queue-all
   *   Clear all migration queues.
   */
  public function migrationClearQueueAll() {
    foreach (array_keys($this->orangeDamMigrationDataManager->getQueues()) as $id) {
      $this->queueFactory->get($id)->deleteQueue();
    }
    $this->displayMigrationQueueListing();
    $this->displayUnregisteredQueueCount();
  }

  /**
   * Clear a specific Orange DAM migration queue.
   *
   * @param array $options
   *   The command options.
   *
   * @command orange-dam:migration-clear-queue
   * @aliases od-mcq
   *
   * @option queue
   *   Allows for the queue to be specified.
   *
   * @usage orange-dam:migration-clear-queue
   *   Clear a specific migration queue.
   */
  public function migrationClearQueue(array $options = [
    'queue' => '',
  ]) {
    $queue = $options['queue'];
    if ($queue) {
      if (!$this->orangeDamMigrationDataManager->validateQueueName($queue)) {
        $this->io()->error(dt(
          'The queue name :queue is an invalid queue name.', [
            ':queue' => $queue,
          ]));
      }
    }
    else {
      $choices = array_keys($this->orangeDamMigrationDataManager->getQueues());
      $choice = $this->io()->choice('Select the queue to clear:', $choices);
      $queue = $choices[$choice];
    }
    $this->loggerChannelFactory
      ->get(OrangeDamQueueDataManager::LOGGER_CHANNEL)
      ->notice(dt('Clearing the %queue queue.', [
        '%queue' => $queue,
      ]));
    $this->queueFactory->get($queue)->deleteQueue();
    $this->displayMigrationQueueListing();
    $this->displayUnregisteredQueueCount();
  }

  /**
   * Provides an overview of migration activity.
   *
   * @param string $time_since
   *   Full date time string in UTC of any PHP parsable format, such as
   *   "1/25/2023 09:47:01".
   * @param string $time_until
   *   Full date time string (in the site's timezone) of any PHP parsable
   *   format, such as "1/25/2023 09:47:01".
   * @param array $options
   *   The command options.
   *
   * @command orange-dam:migration-status
   * @aliases od-ms
   *
   * @options system_id
   *   The system ID from Orange DAM.
   * @options entity_id
   *   The Drupal entity ID.
   * @options digital_id
   *   The Digital ID from Orange DAM.
   *
   * @usage drush orange-dam:migration-status --system_id=EX123 --entity-id=123 --digital-id=EX987
   *
   * @usage orange-dam:migration-status
   *   Provides an overview of the migration activity from the queues since a
   *   given time. Defaults to the past 24 hours if no time is provided.
   *
   * @todo This now uses fields specific to our project. We need to refactor
   *    the field_digital/digital_id filter out to a custom module.
   */
  public function migrationStatus(string $time_since = NULL, string $time_until = NULL, array $options = [
    'system-id' => '',
    'entity-id' => '',
    'digital-id' => '',
  ]) {
    // Set the updated since time, default to 24 hours if no value was provided.
    $time_updated_since = new DrupalDateTime($time_since);
    $time_updated_since->setTimezone(new \DateTimeZone('utc'));
    if (is_null($time_since)) {
      $time_updated_since->modify('-24 hours');
    }
    // Set the updated until time, default to now if no value was provided.
    $time_updated_until = new DrupalDateTime($time_until);
    $time_updated_until->setTimezone(new \DateTimeZone('utc'));
    // Display the filtered migration results.
    $this->displayRecentlyMigratedContent(
      $time_updated_since,
      $time_updated_until,
      $options['system-id'],
      $options['entity-id'],
      $options['digital-id'],
    );
  }

  /**
   * Provides an overview of Orange DAM queues.
   *
   * @command orange-dam:migration-queue-status
   * @aliases od-mqs
   *
   * @usage orange-dam:migration-queue-status
   *   Provides an overview of the sync status of all Orange DAM queues.
   */
  public function migrationQueueStatus() {
    $this->displayMigrationQueueListing();
    $this->displayDeletionQueue();
    $this->displayUnregisteredQueueCount();
  }

  /**
   * Show unregistered queues that lack a corresponding migration.
   *
   * @command orange-dam:migration-missing
   * @aliases od-mm
   *
   * @usage orange-dam:migration-missing
   *   Show unregistered queues that lack a corresponding migration.
   */
  public function migrationMissing() {
    $missing = 0;
    $rows = [];
    $migrations = [];
    $queues_registered = $this->orangeDamMigrationDataManager->getQueues();
    // Get a list of migrations.
    foreach ($this->orangeDamMigrationDataManager->getMigrations() as $migration) {
      $migrations[] = $migration->id();
    }
    // Build a table that shows the migration associated with each queue or
    // denote the queues missing a migration.
    foreach ($this->getOrangeDamQueues() as $queue) {
      $migration_name_from_queue = str_replace('orange_dam_migration:', '', $queue);
      $match = in_array($migration_name_from_queue, $migrations);
      $registered = array_key_exists($queue, $queues_registered);
      if (!$match || !$registered) {
        $missing++;
      }
      $migration_text = $match ? $migration_name_from_queue : '[missing]';
      $registered_text = $registered ? $queues_registered[$queue]['title'] : '[missing]';
      $migration_suggestion = $match ? 'N/A' : $migration_name_from_queue;
      $rows[] = [
        $queue,
        $registered_text,
        $migration_text,
        $migration_suggestion,
      ];
    }
    // Display the results.
    $this->io()->table([
      'Queue Name',
      'Queue Registered',
      'Associated Migration',
      'Suggested Migration ID',
    ], $rows);
    // Display a summary.
    if ($missing > 0) {
      $this->io()->warning(
        $missing . ' queue(s) unregistered or missing a migrations.
        Clearing caches or importing configuration might solve this.
      ');
    }
    else {
      $this->io()->success('All queues are registered and have an associated migration.');
    }
    // Display unregistered queue counts.
    $this->displayUnregisteredQueueCount();
  }

  /**
   * List all available Orange DAM fields.
   *
   * @command orange-dam:migration-list-fields
   * @aliases od-mlf
   *
   * @usage orange-dam:migration-list-fields
   *   Lists all the available Orange DAM fields.
   */
  public function migrationListFields() {
    $rows = [];
    foreach ($this->orangeDamApi->listFields() as $field_id => $field_name) {
      // Abbreviate the field name if necessary.
      if (strlen($field_name) > static::MAX_FIELD_CHARACTER_LENGTH) {
        $field_name = sprintf('%s…',
          strtok(wordwrap($field_name, static::MAX_FIELD_CHARACTER_LENGTH), PHP_EOL)
        );
      }
      // Build the row data.
      $rows[] = [
        $field_id,
        str_replace(PHP_EOL, ' ', $field_name),
      ];
    }
    $this->io()->table(['Field ID', 'Field Name'], $rows);
  }

  /**
   * Generate an asset link.
   *
   * @param string $system_identifier
   *   The system identifier of the item in Orange DAM. Example: EX1STO431061.
   *
   * @command orange-dam:migration-get-asset-link
   * @aliases od-mgal
   *
   * @usage orange-dam:migration-generate-asset-link EX1STO431061>
   *   Generate an asset link.
   */
  public function migrationGetAssetLink(string $system_identifier) {
    try {
      OrangeDamContent::validateSystemIdentifier($system_identifier);
    }
    catch (\Exception $e) {
      $this->loggerChannelFactory->get(OrangeDamMigrationDataManager::LOGGER_CHANNEL)->error($e->getMessage());
      return;
    }
    if ($url = $this->orangeDamUpdatedDataManager->getAssetUrl($system_identifier)) {
      $this->io()->text($url);
    }
  }

  /**
   * Get subtitles for an item.
   *
   * @param string $system_identifier
   *   The system identifier of the item in Orange DAM. Example: EX1STO431061.
   *
   * @command orange-dam:migration-get-subtitles
   * @aliases od-mgs
   *
   * @usage orange-dam:migration-get-subtitles EX1STO431061
   *   Generate an asset link.
   */
  public function migrationSubtitles(string $system_identifier) {
    try {
      OrangeDamContent::validateSystemIdentifier($system_identifier);
    }
    catch (\Exception $e) {
      $this->loggerChannelFactory->get(OrangeDamMigrationDataManager::LOGGER_CHANNEL)->error($e->getMessage());
      return;
    }
    if (!$item = $this->orangeDamApi->getContentItem($system_identifier)) {
      $this->io()->error($this->t(
        'Unable to retrieve an item with system ID ":id".', [
          ':id' => $system_identifier,
        ]));
      return;
    }
    if ($subtitles = $this->orangeDamApi->getSubtitles($item->data()->RecordID)) {
      $this->io()->text(print_r($subtitles));
    }
    else {
      $this->io()->error($this->t(
        'Unable to retrieve subtitles for the item with system ID ":id".', [
          ':id' => $system_identifier,
        ]));
    }
  }

  /**
   * Process the deletion queue.
   *
   * @param string $limit
   *   The maximum number of items to process.
   *
   * @command orange-dam:migration-process-deletion-queue
   * @aliases od-mpdq
   *
   * @usage orange-dam:migration-process-deletion-queue
   *   Process all items in the deletion queue.
   * @usage orange-dam:migration-process-deletion-queue 123
   *   Process 123 items in the deletion queue.
   */
  public function migrationProcessDeletionQueue(string $limit = '') {
    $limit = $limit ?: 0;
    $this->orangeDamMigrationDataManager->processDeletionQueue((int) $limit);
  }

  /**
   * Array sorting function for queue listing reports.
   *
   * @param array $a
   *   The first item to compare.
   * @param array $b
   *   The second item to compare.
   *
   * @return int
   *   Integer indicating sort order.
   */
  public static function sortQueues(array $a, array $b): int {
    return strtolower($a[0]) <=> strtolower($b[0]);
  }

  /**
   * An array of queue names that have the Orange DAM prefix in their name.
   *
   * @return string[]
   *   Orange DAM prefixed queue names.
   */
  protected function getOrangeDamQueues(): array {
    return $this->databaseConnection
      ->select('queue', 'q')
      ->fields('q', ['name'])
      ->condition('name', $this->databaseConnection->escapeLike('orange_dam_migration') . '%', 'LIKE')
      ->distinct()
      ->execute()
      ->fetchCol();
  }

  /**
   * Get an array of unregistered queue names.
   *
   * @return string[]
   *   An array of unregistered queue names.
   */
  protected function getUnregisteredQueueNames(): array {
    $all_queues = $this->getOrangeDamQueues();
    $registered_queues = array_keys($this->orangeDamMigrationDataManager->getQueues());
    return array_diff($all_queues, $registered_queues);
  }

  /**
   * Get the number of items in an unregistered queue by name.
   *
   * @param string $queue_name
   *   The queue name.
   *
   * @return int
   *   The number of items in the queue.
   */
  protected function getQueueItemCount(string $queue_name): int {
    return $this->databaseConnection
      ->select('queue', 'q')
      ->fields('q', ['name'])
      ->condition('name', $queue_name)
      ->countQuery()->execute()->fetchField();
  }

  /**
   * Display a table of all items in unregistered queues.
   */
  protected function displayUnregisteredQueueCount(): void {
    $rows = [];
    $unregistered_queues = $this->getUnregisteredQueueNames();
    if ($unregistered_queues === []) {
      $this->io()->success('No unregistered queue items.');
      return;
    }
    foreach ($unregistered_queues as $unregistered_queue) {
      $rows[] = [
        $unregistered_queue,
        $this->getQueueItemCount($unregistered_queue),
      ];
    }
    $this->io()->title('Unregistered Queue Items');
    $this->io()->table(['Queue Name', 'Count'], $rows);
  }

  /**
   * Display the recently migrated content table.
   *
   * @param \Drupal\Core\Datetime\DrupalDateTime $time_updated_since
   *   The time the migration was run since.
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $time_updated_until
   *   The time the migration was run until.
   * @param string|null $system_id
   *   The System ID from Orange DAM to filter results by.
   * @param string|null $entity_id
   *   The Drupal entity ID to filter results by.
   * @param string|null $digital_id
   *   The Digital ID from Orange DAM to filter results by.
   */
  protected function displayRecentlyMigratedContent(
    DrupalDateTime $time_updated_since,
    DrupalDateTime $time_updated_until,
    ?string $system_id,
    ?string $entity_id,
    ?string $digital_id
  ) {
    // Show the recently migrated items.
    $this->io()->title(dt('Content Migrated Since :time_since to :time_until.', [
      ':time_since' => $time_updated_since->setTimezone(new \DateTimeZone('utc'))->format('c'),
      ':time_until' => $time_updated_until->setTimezone(new \DateTimeZone('utc'))->format('c'),
    ]));
    $rows = [];
    $migrations = $this->orangeDamMigrationDataManager->getMigrations();
    foreach ($migrations as $migration) {
      $table_name = sprintf('migrate_map_%s', $migration->id());
      // Verify that the mapping table exists.
      if (!$this->databaseConnection->schema()->tableExists($table_name)) {
        continue;
      }
      // Find items in the mapping table.
      $query = $this->databaseConnection
        ->select($table_name, 'mm')
        ->fields('mm')
        ->where('last_imported > :time_since', [
          ':time_since' => $time_updated_since->format('U'),
        ])
        ->where('last_imported < :time_until', [
          ':time_until' => $time_updated_until->format('U'),
        ]);
      // Add optional filters if a value was provided.
      if ($system_id) {
        $query->where('mm.sourceid1 = :system_id', [
          ':system_id' => $system_id,
        ]);
      }
      if ($entity_id) {
        $query->where('mm.destid1 = :entity_id', [
          'entity_id' => $entity_id,
        ]);
      }
      if ($digital_id) {
        $query->leftJoin('node', 'n', 'n.nid = mm.destid1');
        $query->leftJoin('node__field_identifier', 'fi', 'fi.revision_id = n.vid');
        $query->where('fi.field_identifier_value = :digital_id', [
          ':digital_id' => $digital_id,
        ]);
      }
      // Run the query and display the results.
      $results = $query->execute()->fetchAll();
      foreach ($results as $result) {
        $last_imported = DrupalDateTime::createFromTimestamp($result->last_imported);
        // Get the entity type and load the entity.
        $entity_type = $migration->getDestinationPlugin()->getDerivativeId();
        $entity = $this->entityTypeManager
          ->getStorage($entity_type)
          ->load($result->destid1);
        if (!$entity) {
          $this->io()->warning(dt('Unable to find the entity with Source ID :source_id and destination ID :destination_id.', [
            ':source_id' => $result->sourceid1,
            ':destination_id' => $result->destid1,
          ]));
        }
        // Get the digital identifier and bundle for the item.
        $digital_identifier = $entity && $entity->hasField('field_identifier') ? $entity->get('field_identifier')->value : 'N/A';
        $bundle = $entity ? $entity->bundle() : 'unknown';
        // Build the title and abbreviate it if necessary.
        $title = $entity ? $entity->label() : 'Unknown Title';
        if (strlen($title) > static::MAX_TITLE_CHARACTER_LENGTH) {
          $title = sprintf('%s…', strtok(wordwrap($title, static::MAX_TITLE_CHARACTER_LENGTH), PHP_EOL));
        }
        // Build the table row.
        $rows[] = [
          $result->sourceid1,
          sprintf('%s:%s', $entity_type, $bundle),
          $result->destid1,
          $digital_identifier,
          $last_imported->setTimezone(new \DateTimeZone('utc'))->format('c'),
          $title,
        ];
      }
    }
    $this->io()->table([
      'System ID',
      'Entity Type',
      'Entity ID',
      'Digital Identifier',
      'Last Updated',
      'Title',
    ], $rows);
  }

  /**
   * Display the migration queue listing table.
   *
   * @param string[] $tags
   *   The array of migration tags to filter the results by.
   */
  protected function displayMigrationQueueListing(array $tags = []): void {
    // Show the size of each migration queue.
    $this->io()->title('Migration Queues');
    $headers = ['Queue Name', 'Items', 'Migration'];
    $rows = [];
    $queues = $this->orangeDamMigrationDataManager->getQueues();
    foreach ($this->orangeDamMigrationDataManager->getMigrations() as $migration) {
      // If tags were supplied, only add rows that match one of the tags.
      if ($tags && !array_intersect($tags, $migration->getMigrationTags())) {
        continue;
      }
      // Build each row.
      $configuration = $migration->getSourceConfiguration();
      $queue_id = $configuration['queue_name'];
      // Alert the user if the queue was not found.
      if (!isset($queues[$queue_id])) {
        $this->io()->error(
          dt('The queue ":queue" is missing or not registered. Create a migration with "id: :queue" to register it.', [
            ':queue' => $queue_id,
          ]));
        continue;
      }
      $rows[] = [
        $queues[$queue_id]['title'],
        number_format($this->queueFactory->get($queue_id)->numberOfItems()),
        $migration->id(),
      ];
    }
    usort($rows, [OrangeDamCommands::class, 'sortQueues']);
    $this->io()->table($headers, $rows);
  }

  /**
   * Display information about the deletion queue.
   */
  protected function displayDeletionQueue() {
    $rows = [];
    // Deletion Queue.
    $this->io()->title('Deletion Queue');
    $deletion_queue = $this->orangeDamMigrationDataManager->getDeletionQueue();
    $rows[] = [
      $this->orangeDamMigrationDataManager::DELETION_QUEUE,
      number_format($deletion_queue->numberOfItems()),
    ];
    $this->io()->table(['Queue Name', 'Items'], $rows);
  }

  /**
   * Display migration results.
   *
   * @param array $results
   *   The results from $this->orangeDamMigrationDataManager->runMigrations.
   */
  protected function displayMigrationResults(array $results) {
    // Warn the user if nothing was run.
    if ($results === []) {
      $this->loggerChannelFactory
        ->get(OrangeDamMigrationDataManager::LOGGER_CHANNEL)
        ->warning(dt(
          'No migrations were run. Ensure that the migrations exist and are tagged with "orange_dam".'
        ));
    }
    // Record the results of the migrations.
    $headers = array_keys($results[0]);
    $headers = array_map('ucwords', $headers);
    $this->io()->table($headers, $results);
  }

  /**
   * Create a date time object while validating the source string format.
   *
   * @param string $time
   *   The date time string.
   *
   * @return \Drupal\Core\Datetime\DrupalDateTime
   *   The date time object.
   *
   * @throws \Exception
   */
  protected function createDateTime(string $time): DrupalDateTime {
    try {
      $date_time = new DrupalDateTime($time);
      $date_time->setTimezone(new \DateTimeZone('utc'));
    }
    catch (\Exception) {
      throw new \Exception(sprintf(
        'Invalid source date format "%s".',
        $time,
      ));
    }
    return $date_time;
  }

}
