<?php

namespace Drupal\orange_dam\EventSubscriber;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\migrate\Event\MigrateEvents;
use Drupal\migrate\Event\MigratePostRowSaveEvent;
use Drupal\migrate\MigrateSkipRowException;
use Drupal\orange_dam\Event\OrangeDamDuplicateItemDeletionEvent;
use Drupal\orange_dam\Event\OrangeDamDuplicateItemDetectionEvent;
use Drupal\orange_dam\Event\OrangeDamDuplicateItemEvents;
use Drupal\orange_dam\OrangeDamApi;
use Drupal\orange_dam\OrangeDamContent;
use Drupal\orange_dam\OrangeDamItemInterface;
use Drupal\orange_dam\OrangeDamKeyword;
use Drupal\orange_dam\OrangeDamMigrationDataManager;
use Drupal\pathauto\PathautoGeneratorInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Checks for duplicate items to handle content type changes.
 */
class OrangeDamMigrationSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * Constructor.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected LoggerChannelFactoryInterface $loggerChannelFactory,
    protected OrangeDamApi $orangeDamApi,
    protected OrangeDamMigrationDataManager $migrationDataManager,
    protected ConfigFactoryInterface $configFactory,
    protected EventDispatcherInterface $eventDispatcher,
    protected PathautoGeneratorInterface $pathautoGenerator,
  ) {
    $this->logger = $loggerChannelFactory->get(OrangeDamMigrationDataManager::LOGGER_CHANNEL);
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    return [
      MigrateEvents::POST_ROW_SAVE => ['checkExistingContent'],
    ];
  }

  /**
   * Handle the post row save checks.
   *
   * @param \Drupal\migrate\Event\MigratePostRowSaveEvent $event
   *   The migration event.
   */
  public function checkExistingContent(MigratePostRowSaveEvent $event): void {
    // Determine if we should check for duplicates.
    if (!$this->isCheckForDuplicatesEnabled()) {
      return;
    }
    if (!$this->isCheckForDuplicatesEnabledForRow($event)) {
      return;
    }
    // Get information about the migrated entity.
    $source_ids = $event->getRow()->getSourceIdValues();
    $destination_ids = $event->getDestinationIdValues();
    $migrated_entity_id = array_pop($destination_ids);
    $entity_type = $event->getMigration()->getDestinationPlugin()->getDerivativeId();
    // Load the migrated entity.
    $migrated_entity = $this->entityTypeManager
      ->getStorage($entity_type)
      ->load($migrated_entity_id);
    // Build an Orange DAM item with the source data.
    $item = new OrangeDamContent(
      array_pop($source_ids),
      (object) $event->getRow()->getSource(),
    );
    // Find any duplicate entities and delete them.
    $entities_to_delete = $this->findDuplicates($item, $migrated_entity);
    foreach ($entities_to_delete as $entity_to_delete) {
      $this->deleteDuplicate($entity_to_delete, $migrated_entity);
    }
    // Log information about the new entity only when duplicates were found.
    if ($entities_to_delete !== []) {
      $this->logger->notice(
        'Creating a new node ":title" (:system_id) of type ":type" because it changed its type in Orange DAM.', [
          ':title' => $migrated_entity->label(),
          ':type' => $migrated_entity->bundle(),
          ':system_id' => $migrated_entity->field_system_id->value,
        ]

      );
    }
    // Update the alias on the new entity now that the alias is removed from the
    // original entity.
    $this->pathautoGenerator->updateEntityAlias($migrated_entity, 'update');
  }

  /**
   * Find existing entities with the same system identifier.
   *
   * @param \Drupal\orange_dam\OrangeDamItemInterface $item
   *   The item to check.
   * @param \Drupal\Core\Entity\EntityInterface $migrated_entity
   *   The entity that was just migrated.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   *   The duplicate entities found.
   */
  protected function findDuplicates(OrangeDamItemInterface $item, EntityInterface $migrated_entity): array {
    /** @var \Drupal\orange_dam\Event\OrangeDamDuplicateItemDetectionEvent $orangeDamDuplicateEvent */
    $orangeDamDuplicateEvent = $this->eventDispatcher->dispatch(
      new OrangeDamDuplicateItemDetectionEvent($item, $migrated_entity),
      OrangeDamDuplicateItemEvents::DETECTED,
    );
    return $orangeDamDuplicateEvent->getDuplicateEntities();
  }

  /**
   * Delete a duplicate entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity_delete
   *   The entity to delete.
   * @param \Drupal\Core\Entity\EntityInterface $entity_new
   *   The new entity that is replacing the duplicate ones.
   */
  protected function deleteDuplicate(EntityInterface $entity_delete, EntityInterface $entity_new): void {
    // Allow other modules to respond to duplicate entity deletion.
    $this->eventDispatcher->dispatch(
      new OrangeDamDuplicateItemDeletionEvent($entity_delete, $entity_new),
      OrangeDamDuplicateItemEvents::DELETED,
    );
    // Delete the old entity.
    $entity_delete->delete();
  }

  /**
   * Determines if we should check for duplicate content items.
   *
   * @return bool
   *   Boolean indicating if we should check for duplicate items.
   */
  protected function isCheckForDuplicatesEnabled(): bool {
    static $check_for_duplicates;
    if ($check_for_duplicates === NULL) {
      // Check the Drupal configuration value.
      $check_for_duplicates = $this->configFactory
        ->get('orange_dam.settings')
        ->get('check_for_duplicates');
      // Allow for state overrides.
      $disable_check_for_duplicates_state = drupal_static(
        OrangeDamMigrationDataManager::DISABLE_DUPLICATE_CHECK_STATE_KEY,
      );
      if ($disable_check_for_duplicates_state === TRUE) {
        $check_for_duplicates = FALSE;
        $this->logger->warning('Duplicate checking disabled via CLI for the current migration process.');
      }
    }
    return $check_for_duplicates;
  }

  /**
   * Allow for duplicate checking to be disabled per various source types.
   *
   * @param \Drupal\migrate\Event\MigratePostRowSaveEvent $event
   *   The migration event.
   *
   * @throws \Drupal\migrate\MigrateSkipRowException
   */
  protected function isCheckForDuplicatesEnabledForRow(MigratePostRowSaveEvent $event): bool {
    $data = $event->getRow()->getSourceProperty('data');
    if (!is_array($data) || !array_key_exists(OrangeDamItemInterface::ITEM_TYPE_PROPERTY, $data)) {
      $message = $this->t(
        'Source data is missing required field :field. Cannot determine whether to check for duplicates.',
        [':field' => OrangeDamItemInterface::ITEM_TYPE_PROPERTY],
      );
      throw new MigrateSkipRowException($message);
    }
    // Keywords do not change their type in Orange DAM to our knowledge.
    return $data[OrangeDamItemInterface::ITEM_TYPE_PROPERTY] !== OrangeDamKeyword::ITEM_TYPE;
  }

}
