<?php

namespace Drupal\orange_dam;

use Chromatic\OrangeDam\DataTableIdentifiers;
use Chromatic\OrangeDam\Endpoints\DataTable;
use Chromatic\OrangeDam\Endpoints\ObjectManagement;
use Chromatic\OrangeDam\Endpoints\Search;
use Chromatic\OrangeDam\Factory as OrangeDamClientFactory;
use Chromatic\OrangeDam\Http\Client;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Utility\Error;
use Drupal\orange_dam\Event\OrangeDamApiRequestEvent;
use Drupal\orange_dam\Event\OrangeDamApiRequestEvents;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\ConnectException;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Output\ConsoleOutput;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Orange DAM API interaction manager.
 */
class OrangeDamApi {

  use StringTranslationTrait;

  /**
   * The oauth expire time state key.
   */
  public const OAUTH_EXPIRE_IN = 'orange_dam.expires_in';

  /**
   * The default API request timeout in seconds.
   */
  public const DEFAULT_API_REQUEST_TIMEOUT = 30;

  /**
   * HTTP status code for OK.
   */
  public const HTTP_STATUS_OK = 200;

  /**
   * Content Types which require DataTable API data.
   *
   * @var array
   */
  protected array $dataTableContentTypes = [];

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $logger;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $currentRequest;

  /**
   * The Orange DAM client.
   *
   * @var \Chromatic\OrangeDam\Http\Client
   */
  protected $orangeDamClient;

  /**
   * Constructor.
   */
  public function __construct(
    protected LoggerChannelFactoryInterface $logger_factory,
    protected ClientInterface $httpClient,
    protected ConfigFactoryInterface $config_factory,
    protected StateInterface $state,
    protected RequestStack $requestStack,
    protected TimeInterface $time,
    protected EventDispatcherInterface $eventDispatcher,
    protected OrangeDamMigrationDataManager $migrationDataManager,
    protected OrangeDamConfigurationManager $orangeDamConfigurationManager,
  ) {
    $this->logger = $logger_factory->get('orange_dam.api');
    $this->config = $config_factory->get('orange_dam.settings');
    $this->currentRequest = $requestStack->getCurrentRequest();
  }

  /**
   * Gets the list of available Orange DAM field names.
   *
   * @return array
   *   An array of Orange DAM field names keyed by the field machine name.
   *
   * @throws \Exception
   */
  public function listFields() {
    $fields = [];
    $params = [
      'format' => 'json',
    ];
    try {
      $response = $this->getOrangeDamClient()->search()->listFields($params);
      $content = $response->getData();
    }
    catch (\Exception $e) {
      $variables = Error::decodeException($e);
      $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
      throw new \Exception($e->getMessage(), $e->getCode(), $e);
    }
    // Validate the response.
    if (empty($content)) {
      throw new \Exception(sprintf(
        'Unexpected Orange DAM API response. Received %s with message "%s."',
        $response->getStatusCode(),
        $response->getReasonPhrase(),
      ));
    }
    // Validate that items were found.
    if (empty($content->APIResponse->Metadata)) {
      $this->logger->error($this->t('No Orange DAM field information found.'));
      return $fields;
    }
    return (array) $content->APIResponse->Metadata;
  }

  /**
   * Retrieves all data about an item from the DataTable API.
   *
   * @param string $identifier
   *   The identifier for the item.
   * @param \Chromatic\OrangeDam\DataTableIdentifiers $identifier_type
   *   The identifier type.
   *
   * @return \Object
   *   The Orange DAM item data. The object will be empty on error.
   */
  public function getDataTableData(string $identifier, DataTableIdentifiers $identifier_type):object {
    // Build the request parameters.
    $params = [
      $identifier_type->value => $identifier,
      'format' => 'json',
    ];
    try {
      $response = $this->getOrangeDamClient()->dataTable()->getDocumentData($params);
      $content = $response->getData();
    }
    catch (\Exception $e) {
      $variables = Error::decodeException($e);
      $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
      throw new \Exception($e->getMessage(), $e->getCode(), $e);
    }
    // Validate the response.
    if (empty($content)) {
      throw new \Exception(sprintf(
        'Unexpected Orange DAM API response. Received %s with message "%s."',
        $response->getStatusCode(),
        $response->getReasonPhrase(),
      ));
    }
    // Log an error if no item with the ID was found.
    if (empty($content->Response[0])) {
      $this->logger->warning('Unable to retrieve Orange DAM DataTable information for item ":id".', [
        ':id' => $identifier,
      ]);
      return new \stdClass();
    }

    return $content->Response[0];
  }

  /**
   * Retrieves an item from the Orange DAM API.
   *
   * The item is also queued.
   *
   * @param string $system_identifier
   *   The System Identifier of the item in Orange DAM. Example: EX1STO431061.
   *
   * @return \Drupal\orange_dam\OrangeDamItemInterface|false
   *   The Orange DAM item.
   */
  public function getContentItem(string $system_identifier) {
    // Build the request parameters.
    $params = [
      'query' => [
        'SystemIdentifier:' . $system_identifier,
      ],
      'fields' => $this->getAllFieldNamesCommaSeparated(),
      'format' => 'json',
    ];
    /** @var \Drupal\orange_dam\Event\OrangeDamApiRequestEvent $orangeDamApiEvent */
    $orangeDamApiEvent = $this->eventDispatcher->dispatch(
      new OrangeDamApiRequestEvent($params),
      OrangeDamApiRequestEvents::ITEM_DATA
    );
    $params = $orangeDamApiEvent->getParameters();
    // Combine the query parameters into a string.
    $params['query'] = implode(' ', $params['query']);
    try {
      $response = $this->getOrangeDamClient()->search()->search($params);
      $content = $response->getData();
    }
    catch (\Exception $e) {
      $variables = Error::decodeException($e);
      $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
      throw new \Exception($e->getMessage(), $e->getCode(), $e);
    }
    // Validate the response.
    if (empty($content)) {
      throw new \Exception(sprintf(
        'Unexpected Orange DAM API response. Received %s with message "%s."',
        $response->getStatusCode(),
        $response->getReasonPhrase(),
      ));
    }
    // Log an error if no item with the ID was found.
    if (empty($content->APIResponse->Items[0])) {
      $this->logger->warning('Unable to retrieve information for item ":id" from Orange DAM.', [
        ':id' => $system_identifier,
      ]);
      return FALSE;
    }
    // Build an Orange DAM item.
    $item = $this->buildOrangeDamContentItem($content->APIResponse->Items[0]);
    // Queue the item and return it.
    $this->migrationDataManager->queueMigrationItem($item);
    return $item;
  }

  /**
   * Retrieves & queues updated items in the Orange DAM API since a given time.
   *
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $time_updated_since
   *   The time to check for items changed since.
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $time_updated_until
   *   The time to check for items changed until.
   * @param array $content_types
   *   The content types to filter the results by. Defaults to all.
   * @param int $limit
   *   The maximum number of items, rounded up to the nearest page size.
   *
   * @return int
   *   The number of Orange DAM items in the API result.
   */
  public function getContent(
    ?DrupalDateTime $time_updated_since,
    ?DrupalDateTime $time_updated_until,
    array $content_types = [],
    int $limit = 0,
  ):int {
    $queued = 0;
    $times = [];
    $quit = FALSE;
    $next_page = FALSE;
    $progress_bar = new ProgressBar(new ConsoleOutput());
    $items_per_page = $this->config->get('api_items_per_page_search');
    // Validate the items per page value.
    if (empty($items_per_page)) {
      $this->logger->warning('The :config config value can not be "0" or empty.', [
        ':config' => 'api_items_per_page_search',
      ]);
    }
    $content_types_filter = $content_types ?: $this->orangeDamConfigurationManager->getContentTypeNames();
    // Build the request parameters.
    $params = [
      'countperpage' => $items_per_page,
      'fields' => $this->getAllFieldNamesCommaSeparated(),
      'format' => 'json',
      'pagenumber' => 1,
      'query' => [],
    ];
    array_walk($content_types_filter, fn(&$x) => $x = "'$x'");
    $params['query'][] = sprintf('DocSubType:(%s)', implode(' OR ', $content_types_filter));
    if ($time_updated_since) {
      $params['query'][] = sprintf('EditDate>:%s', $time_updated_since->format(Search::DATE_FORMAT));
    }
    if ($time_updated_until) {
      $params['query'][] = sprintf('EditDate<:%s', $time_updated_until->format(Search::DATE_FORMAT));
    }
    // Allow other modules to alter the query parameters.
    /** @var \Drupal\orange_dam\Event\OrangeDamApiRequestEvent $orangeDamApiEvent */
    $orangeDamApiEvent = $this->eventDispatcher->dispatch(new OrangeDamApiRequestEvent($params), OrangeDamApiRequestEvents::UPDATED_ITEMS);
    $params = $orangeDamApiEvent->getParameters();
    // Combine the query parameters into a string.
    $params['query'] = implode(' ', $params['query']);
    // If a limit was supplied, log the type and limit parameters used.
    if ($limit) {
      $this->logger->notice(
        'Requesting a maximum of :limit items of type :type.', [
          ':limit' => $limit,
          ':type' => implode(', ', $content_types_filter),
        ]
      );
    }
    // Repeat the request until all items are retrieved.
    do {
      $request_time_before = new DrupalDateTime();
      try {
        $response = $this->getOrangeDamClient()->search()->search($params);
        $content = $response->getData();
      }
      catch (ConnectException $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        $this->logger->warning(
          'An Orange DAM API timeout exception occurred when requesting page :page. Retrying the API request.', [
            ':page' => $params['pagenumber'],
          ]);
        // Retry the request.
        continue;
      }
      catch (\Exception $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        throw new \Exception($e->getMessage(), $e->getCode(), $e);
      }
      $request_time_after = new DrupalDateTime();
      // Validate the response.
      if (empty($content)) {
        throw new \Exception(sprintf(
          'Unexpected Orange DAM API response. Received %s with message "%s."',
          $response->getStatusCode(),
          $response->getReasonPhrase(),
        ));
      }
      // Validate that items were found.
      if (empty($content->APIResponse->Items)) {
        if ($time_updated_since) {
          $this->logger->notice('No Orange DAM items updated since %updated_time.', [
            '%updated_time' => $time_updated_since->setTimezone(new \DateTimeZone('utc'))->format('c'),
          ]);
        }
        if ($content_types) {
          $this->logger->warning('No Orange DAM items of type :type found.', [
            ':type' => implode(', ', $content_types),
          ]);
        }
        return $queued;
      }
      // Configure the progress bar on the first iteration of the loop.
      if (!$next_page) {
        $total_count = $content->APIResponse->GlobalInfo->TotalCount;
        $steps = (int) ceil($total_count / $items_per_page);
        $this->logger->notice(
          'Found :count item(s) of type :type in the Orange Logic Search API results.', [
            ':count' => number_format($total_count),
            ':type' => implode(', ', $content_types_filter),
          ]);
        $progress_bar->setMaxSteps($steps);
        $progress_bar->start();
      }
      // Queue the items for migration.
      foreach ($content->APIResponse->Items as $item) {
        // Check if we have met the limit if one was provided.
        $quit = $limit && $queued >= $limit;
        if ($quit) {
          break;
        }
        // Build the Orange DAM content item and queue it.
        $this->migrationDataManager->queueMigrationItem(
          $this->buildOrangeDamContentItem($item),
        );
        $queued++;
      }
      // Record the elapsed request time.
      $elapsed_time = $request_time_before->diff($request_time_after)->s;
      $times[] = $elapsed_time;
      // Determine if we need to request the next page and increment the pager.
      $next_page = isset($content->APIResponse->GlobalInfo->NextPage);
      $params['pagenumber']++;
      // Update the progress bar.
      $progress_bar->advance();
      // Close things out if we are done.
      if (!$next_page || $quit) {
        $progress_bar->finish();
        $progress_bar->clear();
      }
      if ($quit) {
        break;
      }
    } while ($next_page);
    // Log the average request time if requests were made successfully.
    if ($times !== []) {
      // Log the average request time.
      $this->logger->info('Average API request time: :time seconds.', [
        ':time' => array_sum($times) / count($times),
      ]);
    }
    return $queued;
  }

  /**
   * Generate an asset URL.
   *
   * @param string $system_identifier
   *   The System Identifier of the item in Orange DAM. Example: EX1STO431061.
   * @param string $asset_format
   *   The asset format to use such as "TRX" or "TR1" as defined by Orange DAM.
   *
   * @return string|false
   *   The asset URL or false if there was a problem.
   *
   * @see https://EXAMPLESUBDOMAIN.orangelogic.com/swagger/ui/index#!/AssetLink/APIService_AssetLink_v1_0_CreateFormatLink
   */
  public function getAssetUrl(string $system_identifier, string $asset_format) {
    // Build and send the request.
    $params = [
      'Identifier' => $system_identifier,
      'StickToCurrentVersion' => 'true',
      'LogAccess' => 'true',
      'CreateDownloadLink' => 'false',
      'AssetFormat' => $asset_format,
      'format' => 'json',
    ];
    $attempts = 0;
    // Get the maximum attempts configuration.
    $max_attempts = $this->config->get('api_max_retry_attempts');
    if (empty($max_attempts)) {
      throw new \Exception('Missing Orange DAM Maximum API Retry Attempts ("orange_dam.settings api_max_retry_attempts") configuration.');
    }
    do {
      $attempts++;
      try {
        $response = $this->getOrangeDamClient()->assetLink()->createFormatLink($params);
        $content = $response->getData();
      }
      catch (ConnectException $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        $this->logger->warning(
          'An Orange DAM API timeout exception occurred when requesting the asset URL for item ":id" with the asset format :format. Retrying the API request.', [
            ':id' => $system_identifier,
            ':format' => $asset_format,
          ]);
        // Retry the request.
        continue;
      }
      catch (\Exception $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        throw new \Exception($e->getMessage(), $e->getCode(), $e);
      }
      // Validate the response.
      if (empty($content)) {
        throw new \Exception(sprintf(
          'Unexpected Orange DAM API response. Received %s with message "%s."',
          $response->getStatusCode(),
          $response->getReasonPhrase(),
        ));
      }
      if (!isset($content->APIResponse->Response->Link)) {
        $this->logger->error('Unable to retrieve Orange DAM asset URL data for item :item with the asset format ":format".', [
          ':item' => $system_identifier,
          ':format' => $asset_format,
        ]);
        return FALSE;
      }
      return $content->APIResponse->Response->Link;
    } while ($attempts <= $max_attempts);
    // If it still failed after multiple attempts, log an error.
    $this->logger->error(
      'Giving up on retrieving the asset URL for item :item with format ":format" after :attempts attempts. Orange DAM API timed out after :timeout seconds.', [
        ':attempts' => $max_attempts,
        ':item' => $system_identifier,
        ':format' => $asset_format,
        ':timeout' => $this->getRequestTimeout(),
      ]);
    return FALSE;
  }

  /**
   * Gets and queues all deleted items from the Orange DAM.
   *
   * @param \Drupal\Core\Datetime\DrupalDateTime $time_deleted_since
   *   The time to check for items deleted since.
   * @param \Drupal\Core\Datetime\DrupalDateTime $time_deleted_until
   *   The time to check for items deleted until. Defaults to the current time.
   *
   * @return int|false
   *   The number of items queued for deletion or FALSE if something failed.
   *
   * @throws \Exception
   */
  public function getDeletedContent(
    DrupalDateTime $time_deleted_since,
    DrupalDateTime $time_deleted_until = new DrupalDateTime()
  ) {
    $queued = 0;
    $progress_bar = new ProgressBar(new ConsoleOutput());
    // Build the request parameters.
    $params = [
      'query' => [
        'From' => $time_deleted_since->format(ObjectManagement::DATE_FORMAT),
        'To' => $time_deleted_until->format(ObjectManagement::DATE_FORMAT),
        'Page' => 1,
      ],
    ];
    // Repeat the request until all items are retrieved.
    do {
      try {
        $response = $this->getOrangeDamClient()
          ->ObjectManagement()
          ->listDeletedObjects($params);
        $content = $response->getData();
      }
      catch (\Exception $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        throw new \Exception($e->getMessage(), $e->getCode(), $e);
      }
      // Validate the response.
      if (empty($content)) {
        $this->logger->error('Invalid Orange DAM response data.');
        return $queued;
      }
      // Validate that items were found.
      if (empty($content->deletedObjects)) {
        $this->logger->notice($this->t('No Orange DAM deleted objects found.'));
        return $queued;
      }
      // Configure the progress bar on the first iteration of the loop.
      if (!isset($next_page)) {
        $progress_bar->start();
      }
      // Queue the items for deletion.
      foreach ($content->deletedObjects as $object) {
        $item = new OrangeDamContent(
          $object->identifier,
          $object,
        );
        $this->migrationDataManager->queueDeletedItem($item);
        $queued++;
      }
      // Determine if we need to request the next page and increment the pager.
      $next_page = isset($content->nextPageUrl);
      $params['query']['Page']++;
      // Update the progress bar.
      $progress_bar->advance();
      if (!$next_page) {
        $progress_bar->finish();
        $progress_bar->clear();
      }
    } while ($next_page);
    return $queued;
  }

  /**
   * Retrieves Orange DAM keywords that have been updated since the passed time.
   *
   * The keywords are also queued.
   *
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $time_updated_since
   *   The time to check for items changed since.
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $time_updated_until
   *   The time to check for items changed until. Defaults to the current time.
   * @param string[] $tag_types
   *   The array of tag types to filter the results by.
   *
   * @return int|false
   *   The number of Orange DAM keywords that have been updated or false.
   */
  public function getKeywords(
    ?DrupalDateTime $time_updated_since,
    ?DrupalDateTime $time_updated_until,
    array $tag_types = [],
  ) {
    $queued = 0;
    $total_count = 0;
    $progress_bar = new ProgressBar(new ConsoleOutput());
    $keyword_relationships = $this->getKeywordRelationships();
    $allowed_tag_types = $tag_types ?: $this->config->get('tag_types');
    $items_per_page = $this->config->get('api_items_per_page_keywords');
    // Validate the items per page value.
    if (empty($items_per_page)) {
      $this->logger->warning('The :config config value can not be "0" or empty.', [
        ':config' => 'api_items_per_page_keywords',
      ]);
    }
    // Build the request parameters.
    $params = [
      'format' => 'json',
      'CurrentPage' => 1,
      'ItemsPerPage' => $items_per_page,
      'CoreField.To-be-vetted' => 'false',
    ];
    // Add edit date filters.
    if ($time_updated_since) {
      $params['CoreField.EditDate>'] = $time_updated_since->format(DataTable::DATE_FORMAT);
    }
    if ($time_updated_until) {
      $params['CoreField.EditDate<'] = $time_updated_until->format(DataTable::DATE_FORMAT);
    }
    // Repeat the request until all items are retrieved.
    do {
      try {
        $response = $this->getOrangeDamClient()->dataTable()->listKeywords($params);
        $content = $response->getData();
      }
      catch (\Exception $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        throw new \Exception($e->getMessage(), $e->getCode(), $e);
      }
      // Validate the response.
      if (empty($content)) {
        throw new \Exception(sprintf(
          'Unexpected Orange DAM API response. Received %s with message "%s."',
          $response->getStatusCode(),
          $response->getReasonPhrase(),
        ));
      }
      // Validate that items were found and log an explanation for debugging.
      if (empty($content->Response)) {
        if ($time_updated_since) {
          if ($tag_types !== []) {
            $this->logger->notice(
              'No Orange DAM keywords of type :types updated since :updated_time.', [
                ':types' => implode(', ', $tag_types),
                ':updated_time' => $time_updated_since->setTimezone(new \DateTimeZone('utc'))->format('c'),
              ]);
          }
          else {
            $this->logger->notice(
              'No Orange DAM keywords updated since :updated_time.', [
                ':updated_time' => $time_updated_since->setTimezone(new \DateTimeZone('utc'))->format('c'),
              ]);
          }
        }
        elseif ($tag_types !== []) {
          $this->logger->warning(
            'No Orange DAM keywords of type :types found. Please check your keyword types and try again.', [
              ':types' => implode(', ', $tag_types),
            ]);
        }
        return $queued;
      }
      // Record the unusual behavior to set expectations.
      $this->logger->info(
        'Due to a limitation of the Orange DAM API, we are unable to filter by "Tag-type" in the request and must filter the returned data instead.'
      );
      // Configure the progress bar on the first iteration of the loop.
      if (!isset($next_page)) {
        $total_count = $content->ResponseSummary->TotalItemCount;
        $steps = (int) ceil($total_count / $items_per_page);
        $progress_bar->setMaxSteps($steps);
        $progress_bar->start();
      }
      // Build an array of keywords.
      foreach ($content->Response as $item) {
        // Only queue keywords from white-listed tag types.
        // Adding multiple values to the request parameters is not supported by
        // the DataTable 2.1 or 2.2 API. Thus, we are not able to filter by the
        // Tags.CoreField.KeyType in the request and must use this work-around.
        if (!in_array($item->{'CoreField.Tag-type'}, $allowed_tag_types)) {
          continue;
        }
        // Add the parent keyword ID if a value is found.
        $item->parentKeywordRecordID = $keyword_relationships[$item->RecordID] ?? '';
        // Build the keyword item.
        $keyword = new OrangeDamKeyword(
          $item->RecordID,
          $item,
          $item->{'CoreField.Tag-type'},
        );
        $this->migrationDataManager->queueMigrationItem($keyword);
        $queued++;
      }
      // Determine if we need to request the next page and increment the pager.
      $next_page = isset($content->ResponseSummary->NextPage);
      $params['CurrentPage']++;
      // Update the progress bar.
      $progress_bar->advance();
      if (!$next_page) {
        $progress_bar->finish();
        $progress_bar->clear();
      }
    } while ($next_page);
    // Display a final summary.
    $this->logger->notice(
      'Filtering the results by the allowed tag types removed :count keywords.', [
        ':count' => $total_count - $queued,
      ]);
    return $queued;
  }

  /**
   * Get subtitles for an item.
   *
   * @param string $record_id
   *   The item's Orange Logic Record ID.
   *
   * @return object|false
   *   The Orange Logic subtitle data object, or false on error.
   *
   *   The "content" property in the returned value is the primary content.
   */
  public function getSubtitles(string $record_id) {
    $attempts = 0;
    // Get the maximum attempts configuration.
    $max_attempts = $this->config->get('api_max_retry_attempts');
    if (empty($max_attempts)) {
      throw new \Exception('Missing Orange DAM Maximum API Retry Attempts ("orange_dam.settings api_max_retry_attempts") configuration.');
    }
    do {
      $attempts++;
      try {
        $response = $this->getOrangeDamClient()->mediaFile()->getCaptions($record_id);
        $content = $response->getData();
      }
      catch (ConnectException $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        $this->logger->warning(
          'An Orange DAM API connection exception occurred when requesting the media file captions for item ":id". Retrying the API request.', [
            ':id' => $record_id,
          ]);
        // Retry the request.
        continue;
      }
      catch (\Exception $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        throw new \Exception($e->getMessage(), $e->getCode(), $e);
      }
      // Validate the response.
      if ($response->getStatusCode() != static::HTTP_STATUS_OK) {
        throw new \Exception(sprintf(
          'Unexpected Orange DAM API response. Received %s with message "%s."',
          $response->getStatusCode(),
          $response->getReasonPhrase(),
        ));
      }
      if (empty($content) || !isset($content[0]->content)) {
        $this->logger->debug('Empty caption data for media file with record ID :item.', [
          ':item' => $record_id,
        ]);
        return FALSE;
      }
      return $content[0];
    } while ($attempts <= $max_attempts);
    // If it still failed after multiple attempts, log an error.
    $this->logger->error(
      'Giving up on retrieving the subtitles for item :item after :attempts attempts. Orange DAM API timed out after :timeout seconds.', [
        ':attempts' => $max_attempts,
        ':item' => $record_id,
        ':timeout' => $this->getRequestTimeout(),
      ]);
    return FALSE;
  }

  /**
   * Get and queue the deleted keywords.
   *
   * @return int|false
   *   The number of items queued for deletion or FALSE if something failed.
   */
  public function getDeletedKeywords(
    DrupalDateTime $time_deleted_since,
    DrupalDateTime $time_deleted_until = new DrupalDateTime()
  ) {
    $queued = 0;
    $total_count = 0;
    $progress_bar = new ProgressBar(new ConsoleOutput());
    $synced_tag_types = $this->config->get('tag_types');
    $items_per_page = $this->config->get('api_items_per_page_keywords');
    // Validate the items per page value.
    if (empty($items_per_page)) {
      $this->logger->warning('The :config config value can not be "0" or empty.', [
        ':config' => 'api_items_per_page_keywords',
      ]);
    }
    // Build the request parameters.
    $params = [
      'format' => 'json',
      'CurrentPage' => 1,
      'ItemsPerPage' => $items_per_page,
      'Tags.CoreField.EditDate>' => $time_deleted_since->format(ObjectManagement::DATE_FORMAT),
      'Tags.CoreField.EditDate<' => $time_deleted_until->format(ObjectManagement::DATE_FORMAT),
      'Tags.CoreField.Disabled' => 1,
    ];
    // Repeat the request until all items are retrieved.
    do {
      try {
        $response = $this->getOrangeDamClient()->dataTable()->listKeywordDeletions($params);
        $content = $response->getData();
      }
      catch (\Exception $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        throw new \Exception($e->getMessage(), $e->getCode(), $e);
      }
      // Validate the response.
      if (empty($content)) {
        throw new \Exception(sprintf(
          'Unexpected Orange DAM API response. Received %s with message "%s."',
          $response->getStatusCode(),
          $response->getReasonPhrase(),
        ));
      }
      // Build a progress bar.
      if (!isset($next_page)) {
        $total_count = $content->ResponseSummary->TotalItemCount;
        $steps = (int) ceil($total_count / $items_per_page);
        $this->logger->notice('Retrieving :count deleted keywords(s) from Orange Logic.', [
          ':count' => number_format($total_count),
        ]);
        $progress_bar->setMaxSteps($steps);
        $progress_bar->start();
      }
      // Build an array of deleted keywords.
      foreach ($content->Response as $item) {
        // Only queue keywords from white-listed tag types.
        // Adding multiple values to the request parameters is not supported by
        // the DataTable 2.1 or 2.2 API. Thus, we are not able to filter by the
        // Tags.CoreField.KeyType in the request and must use this work-around.
        $type = $item->{'Tags.CoreField.KeyType'};
        if (!in_array($type, $synced_tag_types)) {
          continue;
        }
        $keyword = new OrangeDamKeyword(
          $item->RecordID,
          $item,
          $type,
        );
        $this->migrationDataManager->queueDeletedItem($keyword);
        $queued++;
      }
      // Determine if we need to request the next page and increment the pager.
      $next_page = isset($content->ResponseSummary->NextPage);
      $params['CurrentPage']++;
      // Update the progress bar.
      $progress_bar->advance();
      if (!$next_page) {
        $progress_bar->finish();
        $progress_bar->clear();
      }
    } while ($next_page);

    // Display a final summary.
    $this->logger->notice(
      'Filtering the results by the allowed tag types removed :count keywords.', [
        ':count' => $total_count - $queued,
      ]);
    $this->logger->notice(
      'Queued deletions for :count Orange DAM keywords.', [
        ':count' => $queued,
      ]);
    return $queued;
  }

  /**
   * Get the keyword relationship data from Orange DAM.
   *
   * @return string[]
   *   An array of parent keyword ID data keyed by keyword ID.
   */
  protected function getKeywordRelationships() {
    $keyword_relationships = [];
    // Build the request parameters.
    $params = [
      'format' => 'json',
      'CurrentPage' => 1,
      'ItemsPerPage' => $this->config->get('api_items_per_page_keywords'),
    ];
    // Repeat the request until all items are retrieved.
    do {
      try {
        $response = $this->getOrangeDamClient()->dataTable()->listKeywordRelationships($params);
        $content = $response->getData();
      }
      catch (\Exception $e) {
        $variables = Error::decodeException($e);
        $this->logger->error('%type: @message in %function (line %line of %file).', $variables);
        throw new \Exception($e->getMessage(), $e->getCode(), $e);
      }
      // Validate the response.
      if (empty($content)) {
        throw new \Exception(sprintf(
          'Unexpected Orange DAM API response. Received %s with message "%s."',
          $response->getStatusCode(),
          $response->getReasonPhrase(),
        ));
      }
      // Build an array of keyword relationship data..
      foreach ($content->Response as $item) {
        $keyword_relationships[$item->RecordID] = $item->{'CoreField.KeyidFather'};
      }
      // Determine if we need to request the next page and increment the pager.
      $next_page = isset($content->ResponseSummary->NextPage);
      $params['CurrentPage']++;
    } while ($next_page);
    return $keyword_relationships;
  }

  /**
   * Get the full list of field names to retrieve.
   *
   * @return string[]
   *   An array of all field names to retrieve.
   */
  protected function getAllFieldNames() {
    // @todo Consolidate these configurations if they are not used separately.
    // @todo Determine if an error should be thrown if this configuration is
    //   missing or if we should just rely upon the API to surface the issue.
    return array_unique(array_merge(
      $this->config->get('all_fields'),
      $this->config->get('base_fields')
    ));
  }

  /**
   * Get a comma separated string of all field names to retrieve.
   *
   * @return string
   *   A comma separated string of all field names to retrieve.
   */
  protected function getAllFieldNamesCommaSeparated(): string {
    return implode(',', $this->getAllFieldNames());
  }

  /**
   * Check if the Orange DAM API is configured.
   *
   * When orangeDam is configured, the refresh token will be set.
   *
   * @return bool
   *   True if the OAuth token is set. False, otherwise.
   */
  protected function isConfigured(): bool {
    return !empty($this->orangeDamClient->getClient()->token);
  }

  /**
   * Initialize and return the Orange DAM client instance.
   *
   * @return \Chromatic\OrangeDam\Factory
   *   Orange DAM API client instance.
   */
  protected function getOrangeDamClient(): OrangeDamClientFactory {
    if (!$this->orangeDamClient instanceof Client) {
      // Validate that the needed configuration is available.
      $base_path = $this->config->get('base_path');
      if (empty($base_path)) {
        throw new \Exception('Missing Orange DAM Base Path ("orange_dam.settings base_path") configuration.');
      }
      // Build the client.
      $this->orangeDamClient = new OrangeDamClientFactory(
        [
          'base_path' => $base_path,
          'query_string' => $this->config->get('query_string'),
        ],
        NULL,
        [
          'http_errors' => FALSE,
          'timeout' => $this->getRequestTimeout(),
        ]
      );
    }
    // Update the OAuth token if it is expired.
    if (!$this->isConfigured() || $this->state->get(static::OAUTH_EXPIRE_IN) < $this->time->getRequestTime()) {
      // Validate that the needed configuration is available.
      if (empty($this->config->get('client_id'))) {
        throw new \Exception('Missing Orange DAM Client ID ("orange_dam.settings client_id") configuration.');
      }
      if (empty($this->config->get('client_secret'))) {
        throw new \Exception('Missing Orange DAM Client Secret ("orange_dam.settings client_secret") configuration.');
      }
      // Generate the tokens.
      $tokens = $this->orangeDamClient->oAuth2()->getTokensByCode(
        $this->config->get('client_id'),
        $this->config->get('client_secret')
      );
      // Adds the OAuth token to the request.
      $this->orangeDamClient->getClient()->setOauth2Token($tokens->access_token);
      // Store the expires in time.
      $this->state->set(
        static::OAUTH_EXPIRE_IN,
        ($tokens->expires_in + $this->currentRequest->server->get('REQUEST_TIME'))
      );
    }
    return $this->orangeDamClient;
  }

  /**
   * Build an Orange DAM Content object from the Search API response.
   *
   * @param object $api_item
   *   The content item as returned from the Search API.
   */
  protected function buildOrangeDamContentItem(object $api_item): OrangeDamContent {
    $system_identifier = $api_item->SystemIdentifier;
    $content_type = $api_item->{OrangeDamContent::CONTENT_TYPE_PROPERTY};
    $data = $api_item;
    if ($this->includeDataTableData($content_type)) {
      $data = (object) array_merge(
        (array) $api_item,
        (array) $this->getDataTableData($system_identifier, DataTableIdentifiers::SYSTEM_ID),
      );
    }
    return new OrangeDamContent(
      $system_identifier,
      $data,
      $content_type,
    );
  }

  /**
   * Get the default API request timeout.
   *
   * @return int
   *   The request timeout in seconds.
   */
  protected function getRequestTimeout(): int {
    return (int) $this->config->get('api_request_timeout') ?: static::DEFAULT_API_REQUEST_TIMEOUT;
  }

  /**
   * Determine if DataTable data should be added to the content item.
   *
   * @param string $content_type
   *   The content type.
   */
  protected function includeDataTableData(string $content_type): bool {
    if (!array_key_exists($content_type, $this->dataTableContentTypes)) {
      $content_types = $this->orangeDamConfigurationManager->getContentTypes();
      $this->dataTableContentTypes[$content_type] = array_key_exists($content_type, $content_types)
        && array_key_exists('use_datatable', $content_types[$content_type])
        && $content_types[$content_type]['use_datatable'];
    }
    return $this->dataTableContentTypes[$content_type];
  }

}
