<?php

namespace Drupal\orange_dam;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\State;

/**
 * Manages retrieving a list of updated content and building a queue.
 */
class OrangeDamQueueDataManager implements OrangeDamQueueDataManagerInterface {

  /**
   * The log channel to use.
   */
  public const LOGGER_CHANNEL = 'orange_dam.queue';

  /**
   * The state key that tracks the last time we checked for updates to content.
   */
  public const UPDATED_ITEM_STATE_TIME_KEY = 'orange_dam_item_update_last_run_time';

  /**
   * The state key that tracks the last time we checked for updates to keywords.
   */
  public const UPDATED_KEYWORD_STATE_TIME_KEY = 'orange_dam_keyword_update_last_run_time';

  /**
   * The state key that tracks the last time we checked for deleted content.
   */
  public const DELETED_CONTENT_STATE_TIME_KEY = 'orange_dam_deleted_content_last_run_time';

  /**
   * The state key that tracks the last time we checked for deleted keywords.
   */
  public const DELETED_KEYWORD_STATE_TIME_KEY = 'orange_dam_deleted_keyword_last_run_time';

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The Orange DAM configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Constructor.
   */
  public function __construct(
    protected LoggerChannelFactoryInterface $loggerChannelFactory,
    protected QueueFactory $queueFactory,
    protected OrangeDamApi $orangeDamApi,
    protected State $state,
    protected OrangeDamMigrationDataManager $migrationDataManager,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected OrangeDamConfigurationManager $orangeDamConfiguration,
  ) {
    $this->logger = $loggerChannelFactory->get(static::LOGGER_CHANNEL);
  }

  /**
   * Queue all updated keywords and content.
   */
  public function syncUpdates() {
    $this->queueKeywordDeletions();
    $this->queueContentDeletions();
    $this->queueUpdatedKeywords();
    $this->queueUpdatedItems();
  }

  /**
   * The log for queue related issues.
   *
   * @return \Drupal\Core\Logger\LoggerChannelInterface
   *   The logger for this service.
   */
  public function logger() {
    return $this->logger;
  }

  /**
   * {@inheritdoc}
   */
  public function queueUpdatedItems(DrupalDateTime $time_updated_since = NULL) {
    // Use the passed time if available. Fallback to the updated time state
    // value, which defaults to now if this is the first run.
    if (is_null($time_updated_since)) {
      $time_updated_since = new DrupalDateTime($this->state->get(static::UPDATED_ITEM_STATE_TIME_KEY, 'now'));
    }
    $time = $time_updated_since->setTimezone(new \DateTimeZone('utc'))->format('c');
    $this->logger->notice(
      'Queueing all Orange DAM content for re-syncing with changes since :time.', [
        ':time' => $time,
      ]);
    // Record the last time this was run for future checks.
    $time_now = new DrupalDateTime();
    $time_now->setTimezone(new \DateTimeZone('utc'));
    // Get the items deleted since the given time until now.
    $item_count = $this->orangeDamApi->getContent($time_updated_since, $time_now);
    if ($item_count === FALSE) {
      return FALSE;
    }
    // Record the last time this was run for future checks.
    $this->state->set(static::UPDATED_ITEM_STATE_TIME_KEY, $time_now->format('c'));
    // Log the results.
    $this->logger->notice(
      'Queued :count Orange DAM content item(s) for re-syncing with changes after :time.', [
        ':time' => $time,
        ':count' => $item_count,
      ]);
    return $item_count;
  }

  /**
   * {@inheritdoc}
   */
  public function queueUpdatedKeywords(DrupalDateTime $time_updated_since = NULL) {
    // Use the passed time if available. Fallback to the updated time state
    // value, which defaults to now if this is the first run.
    if (is_null($time_updated_since)) {
      $time_updated_since = new DrupalDateTime($this->state->get(static::UPDATED_KEYWORD_STATE_TIME_KEY, 'now'));
    }
    // Standardize the timezone.
    $time_updated_since->setTimezone(new \DateTimeZone('utc'));
    // Record the time of the request.
    $time_now = new DrupalDateTime();
    $time_now->setTimezone(new \DateTimeZone('utc'));
    // Log the action being taken.
    $this->logger->notice(
      'Queueing all Orange DAM keywords for re-syncing that have been updated since :time.', [
        ':time' => $time_updated_since->format('c'),
      ]);
    // Get the items updated since the given time until now.
    $updated_keyword_count = $this->orangeDamApi->getKeywords($time_updated_since, $time_now);
    if ($updated_keyword_count === FALSE) {
      return FALSE;
    }
    // Record the last time this was run for future checks.
    $this->state->set(static::UPDATED_KEYWORD_STATE_TIME_KEY, $time_now->format('c'));
    // Log the results.
    $this->logger->notice(
      'Queued :count Orange DAM keyword(s) for re-syncing.', [
        ':count' => $updated_keyword_count,
      ]);
    return $updated_keyword_count;
  }

  /**
   * Queue updates for all items of a given keyword type.
   *
   * @param string[] $keyword_types
   *   The keyword types to queue updates for.
   *
   * @return int
   *   The number of queued keywords.
   */
  public function queueKeywordByType(array $keyword_types): int {
    // Validate the keyword types.
    foreach ($keyword_types as $keyword_type) {
      if (!$this->validateKeywordType($keyword_type)) {
        $this->logger->error('Invalid keyword type :type requested for syncing.', [
          ':type' => $keyword_type,
        ]);
      }
    }
    // Request the data from the API and queue the results.
    $updated_keyword_count = $this->orangeDamApi->getKeywords(
      NULL,
      NULL,
      $keyword_types,
    );
    // Log the results.
    $this->logger->notice(
      'Queueing :count Orange DAM :type keywords(s) for re-syncing.', [
        ':count' => $updated_keyword_count,
        ':type' => implode(', ', $keyword_types),
      ]);
    return $updated_keyword_count;
  }

  /**
   * {@inheritdoc}
   */
  public function updateSingleItem(string $system_identifier) {
    if ($item = $this->orangeDamApi->getContentItem($system_identifier)) {
      $this->logger->notice(
        'Orange DAM item :system_identifier has been successfully queued for updating in the :queue queue.', [
          ':system_identifier' => $system_identifier,
          ':queue' => $this->migrationDataManager->getItemQueueName($item),
        ]);
      return $item;
    }
    $this->logger->error(
      'Failed to retrieve Orange DAM item :system_id from the Orange DAM API.', [
        ':system_id' => $system_identifier,
      ]
    );
    return FALSE;
  }

  /**
   * Queue updates for all items of a given content type.
   *
   * @param array $content_types
   *   The content types to queue updates for.
   * @param int $limit
   *   The maximum number of items, rounded to the nearest page size.
   */
  public function queueContentByType(array $content_types, int $limit = 0) {
    $item_count = 0;
    // Validate the content types.
    foreach ($content_types as $content_type) {
      if (!$this->validateContentType($content_type)) {
        $this->logger->error(
          'Invalid content type :type requested for syncing.', [
            ':type' => $content_type,
          ]);
        continue;
      }
      // Request the data from the API and queue the results.
      $item_count += $this->orangeDamApi->getContent(
        NULL,
        NULL,
        [$content_type],
        $limit,
      );
    }
    // Record the total number of items queued.
    $this->logger->notice(
      'Queued :count Orange DAM :types item(s) for re-syncing.', [
        ':count' => $item_count,
        ':types' => implode(', ', $content_types),
      ]);
  }

  /**
   * {@inheritdoc}
   */
  public function queueContentDeletions(DrupalDateTime $time_deleted_since = NULL) {
    // Use the passed time if available. Fallback to the deleted time state
    // value, which defaults to now if this is the first run.
    if (is_null($time_deleted_since)) {
      $time_deleted_since = new DrupalDateTime($this->state->get(static::DELETED_CONTENT_STATE_TIME_KEY, 'now'));
    }
    // Record the last time this was run for future checks.
    $time_now = new DrupalDateTime();
    $time_now->setTimezone(new \DateTimeZone('utc'));
    // Log the action being taken.
    $this->logger->notice(
      'Syncing deletions for all Orange DAM content that has been deleted since :time.', [
        ':time' => $time_deleted_since->setTimezone(new \DateTimeZone('utc'))->format('c'),
      ]);
    // Get the items updated since the given time until now.
    $item_count = $this->orangeDamApi->getDeletedContent($time_deleted_since, $time_now);
    if ($item_count === FALSE) {
      return FALSE;
    }
    // Record the time since the request succeeded.
    $this->state->set(static::DELETED_CONTENT_STATE_TIME_KEY, $time_now->format('c'));
    // Display a final summary.
    $this->logger->notice(
      'Queued deletions for :count Orange DAM content items.', [
        ':count' => $item_count,
      ]);
    return $item_count;
  }

  /**
   * {@inheritdoc}
   */
  public function queueKeywordDeletions(DrupalDateTime $time_deleted_since = NULL) {
    // Use the passed time if available. Fallback to the deleted time state
    // value, which defaults to now if this is the first run.
    if (is_null($time_deleted_since)) {
      $time_deleted_since = new DrupalDateTime($this->state->get(static::DELETED_KEYWORD_STATE_TIME_KEY, 'now'));
    }
    // Record the last time this was run for future checks.
    $time_now = new DrupalDateTime();
    $time_now->setTimezone(new \DateTimeZone('utc'));

    $this->logger->notice(
      'Syncing deletions for all Orange DAM keywords that have been deleted since :time.', [
        ':time' => $time_deleted_since->setTimezone(new \DateTimeZone('utc'))->format('c'),
      ]);
    // Get the items updated since the given time until now.
    $item_count = $this->orangeDamApi->getDeletedKeywords($time_deleted_since);
    if ($item_count === FALSE) {
      return FALSE;
    }
    // Record the time since the request succeeded.
    $this->state->set(static::DELETED_KEYWORD_STATE_TIME_KEY, $time_now->format('c'));
    return $item_count;
  }

  /**
   * Deletes an Orange DAM item during queue processing.
   *
   * @param \Drupal\orange_dam\OrangeDamItemInterface $item
   *   The Orange DAM item to use as the source of the deletion operation.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *
   * @see \Drupal\orange_dam\Plugin\QueueWorker\OrangeDamDeletionQueueWorker
   */
  public function processDeletedQueueItem(OrangeDamItemInterface $item) {
    // Attempt to find the item in a migration.
    if (!$migration = $this->migrationDataManager->getMigrationBySourceId($item->id())) {
      $this->logger->error('Failed to delete an item of type ":type" while processing the deletion queue. Unable to find a migration with ":id" as a source identifier.', [
        ':id' => $item->id(),
        ':type' => $item->type(),
      ]);
      return;
    }
    // Get the destination entity ID and type values.
    $entity_ids = $migration->getIdMap()->lookupDestinationIds([$item->id()]);
    $entity_ids = array_pop($entity_ids);
    $entity_type = $migration->getDestinationPlugin()->getDerivativeId();
    // Load the entities.
    $entities = $this->entityTypeManager
      ->getStorage($entity_type)
      ->loadMultiple($entity_ids);
    // Validate that entities were found.
    if ($entities === []) {
      $this->logger->error('Failed to find an entity with the ID :id of the Orange Logic type ":type" while processing deletions for item ":source_id".', [
        ':id' => implode(', ', $entity_ids),
        ':type' => $item->type(),
        ':source_id' => $item->id(),
      ]);
      return;
    }
    // Delete the entities.
    foreach ($entities as $entity) {
      try {
        $entity->delete();
        $this->logger->info('Deleted entity (:type::id) ":title" during Orange DAM deletion syncing.', [
          ':type' => $entity->getEntityType()->id(),
          ':id' => $entity->id(),
          ':title' => $entity->label(),
        ]);
      }
      catch (EntityStorageException) {
        $this->logger->error('Failed to delete entity (:type::id) ":title" during Orange DAM deletion syncing.', [
          ':type' => $entity->getEntityType(),
          ':id' => $entity->id(),
          ':title' => $entity->label(),
        ]);
        continue;
      }
    }
  }

  /**
   * Validate if a content type is valid.
   *
   * @param string $content_type
   *   The content type name.
   *
   * @return bool
   *   Indication if the content type is valid.
   */
  public function validateContentType($content_type): bool {
    return in_array($content_type, $this->orangeDamConfiguration->getContentTypeNames());
  }

  /**
   * Validate if a keyword type is valid.
   *
   * @param string $keyword_type
   *   The keyword type name.
   *
   * @return bool
   *   Indication if the keyword type is valid.
   */
  public function validateKeywordType($keyword_type): bool {
    return in_array($keyword_type, $this->orangeDamConfiguration->getKeywordTypes());
  }

  /**
   * Generate the URL to a file asset for a given item.
   *
   * The content type is optional, but if it is omitted it adds a second API
   * call to this process.
   *
   * @param string $system_id
   *   The item's System ID.
   * @param string $content_type
   *   The item's content type.
   *
   * @return string|false
   *   The asset URL if one was generated.
   */
  public function getAssetUrl(string $system_id, string $content_type = NULL) {
    // Get the content type of item if it was not passed in.
    if (is_null($content_type)) {
      $item = $this->orangeDamApi->getContentItem($system_id);
      $content_type = $item ? $item->type() : NULL;
    }
    // Confirm that a content type was found.
    if (!$content_type) {
      $this->logger->error(
        'Unable to determine the content type of the item with system ID :system_id for asset URL generation.', [
          ':system_id' => $system_id,
        ]);
      return FALSE;
    }
    // Get the asset format and verify that a value was found.
    $asset_format = $this->getAssetFormat($content_type);
    if (!$asset_format) {
      return FALSE;
    }
    // Generate the asset URL.
    return $this->orangeDamApi->getAssetUrl(
      $system_id,
      $asset_format,
    );
  }

  /**
   * Determine the asset format to use for a given content type.
   *
   * @param string $content_type
   *   The content type.
   *
   * @return string|false
   *   The asset URL or FALSE if one was unable to be generated.
   *
   * @throws \Exception
   */
  public function getAssetFormat(string $content_type) {
    $asset_formats = $this->orangeDamConfiguration->getAssetFormats();
    // Map the content type to the asset format.
    foreach ($asset_formats as $asset_format => $content_types) {
      if (in_array($content_type, $content_types)) {
        return $asset_format;
      }
    }
    // Log a warning if no match was found.
    $this->logger->warning(
      'Unable to find an asset type mapping for content type :content_type.', [
        ':content_type' => $content_type,
      ]);
    return FALSE;
  }

}
